import datetime
from typing import Any, Dict, List, Type, TypeVar

import attr
from dateutil.parser import isoparse

T = TypeVar("T", bound="ComfyWorkflowItem")


@attr.s(auto_attribs=True)
class ComfyWorkflowItem:
    """
    Attributes:
        name (str):
        title (str):
        user_id (str):
        created_at (datetime.datetime):
        user_nickname (str):
    """

    name: str
    title: str
    user_id: str
    created_at: datetime.datetime
    user_nickname: str
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        title = self.title
        user_id = self.user_id
        created_at = self.created_at.isoformat()

        user_nickname = self.user_nickname

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "title": title,
                "user_id": user_id,
                "created_at": created_at,
                "user_nickname": user_nickname,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        title = d.pop("title")

        user_id = d.pop("user_id")

        created_at = isoparse(d.pop("created_at"))

        user_nickname = d.pop("user_nickname")

        comfy_workflow_item = cls(
            name=name,
            title=title,
            user_id=user_id,
            created_at=created_at,
            user_nickname=user_nickname,
        )

        comfy_workflow_item.additional_properties = d
        return comfy_workflow_item

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from .parser import FalClientParser


def _list(args):
    import posixpath

    from fal.files import FalFileSystem

    fs = FalFileSystem()

    for entry in fs.ls(args.path, detail=True):
        name = posixpath.basename(entry["name"])
        color = "blue" if entry["type"] == "directory" else "default"
        args.console.print(f"[{color}]{name}[/{color}]")


def _download(args):
    from fal.files import FalFileSystem

    fs = FalFileSystem()
    fs.get(args.remote_path, args.local_path, recursive=True)


def _upload(args):
    from fal.files import FalFileSystem

    fs = FalFileSystem()
    fs.put(args.local_path, args.remote_path, recursive=True)


def _upload_url(args):
    from fal.files import FalFileSystem

    fs = FalFileSystem()
    fs.put_file_from_url(args.url, args.remote_path)


def _mv(args):
    from fal.files import FalFileSystem

    fs = FalFileSystem()
    fs.mv(args.source, args.destination)


def add_parser(main_subparsers, parents):
    files_help = "Manage fal files."
    parser = main_subparsers.add_parser(
        "files",
        aliases=["file"],
        description=files_help,
        help=files_help,
        parents=parents,
    )

    subparsers = parser.add_subparsers(
        title="Commands",
        metavar="command",
        dest="cmd",
        required=True,
        parser_class=FalClientParser,
    )

    list_help = "List files."
    list_parser = subparsers.add_parser(
        "list",
        aliases=["ls"],
        description=list_help,
        help=list_help,
        parents=parents,
    )
    list_parser.add_argument(
        "path",
        nargs="?",
        type=str,
        help="The path to list",
        default="/",
    )
    list_parser.set_defaults(func=_list)

    download_help = "Download files."
    download_parser = subparsers.add_parser(
        "download",
        description=download_help,
        help=download_help,
        parents=parents,
    )
    download_parser.add_argument(
        "remote_path", type=str, help="Remote path to download"
    )
    download_parser.add_argument(
        "local_path", type=str, help="Local path to download to"
    )
    download_parser.set_defaults(func=_download)

    upload_help = "Upload files."
    upload_parser = subparsers.add_parser(
        "upload",
        description=upload_help,
        help=upload_help,
        parents=parents,
    )
    upload_parser.add_argument("local_path", type=str, help="Local path to upload")
    upload_parser.add_argument("remote_path", type=str, help="Remote path to upload to")
    upload_parser.set_defaults(func=_upload)

    upload_url_help = "Upload file from URL."
    upload_url_parser = subparsers.add_parser(
        "upload-url",
        description=upload_url_help,
        help=upload_url_help,
        parents=parents,
    )
    upload_url_parser.add_argument("url", type=str, help="URL to upload")
    upload_url_parser.add_argument(
        "remote_path", type=str, help="Remote path to upload to"
    )
    upload_url_parser.set_defaults(func=_upload_url)

    mv_help = "Move or rename a remote file or directory."
    mv_parser = subparsers.add_parser(
        "mv",
        description=mv_help,
        help=mv_help,
        parents=parents,
    )
    mv_parser.add_argument("source", type=str, help="Remote source path")
    mv_parser.add_argument("destination", type=str, help="Remote destination path")
    mv_parser.set_defaults(func=_mv)



from __future__ import annotations

import os
from contextlib import contextmanager
from typing import Dict, Iterator, List, Optional

SETTINGS_SECTION = "__internal__"

NO_PROFILE_ERROR = ValueError(
    "No profile set. Use command [bold]fal profile set[/] to set a profile."
)


class Config:
    _config: Dict[str, Dict[str, str]]
    _profile: Optional[str]
    _editing: bool = False

    DEFAULT_CONFIG_PATH = "~/.fal/config.toml"

    def __init__(self, *, validate_profile: bool = False):
        import tomli

        self.config_path = os.path.expanduser(
            os.getenv("FAL_CONFIG_PATH", self.DEFAULT_CONFIG_PATH)
        )

        try:
            with open(self.config_path, "rb") as file:
                self._config = tomli.load(file)
        except FileNotFoundError:
            self._config = {}

        profile = os.getenv("FAL_PROFILE") or self.get_internal("profile")

        # Try to set the profile, but don't fail if it doesn't exist
        try:
            self.profile = profile
        except ValueError:
            # Profile doesn't exist, set to None
            self._profile = None

        if validate_profile and not self.profile:
            raise NO_PROFILE_ERROR

    @property
    def profile(self) -> Optional[str]:
        return self._profile

    @profile.setter
    def profile(self, value: Optional[str]) -> None:
        if value and value not in self._config:
            # Don't automatically create profiles - they should be created explicitly
            raise ValueError(
                f"Profile '{value}' does not exist. Create it first or use the profile set command."  # noqa: E501
            )
        elif not value:
            self.unset_internal("profile")

        self._profile = value

    def profiles(self) -> List[str]:
        keys: List[str] = []
        for key in self._config:
            if key != SETTINGS_SECTION:
                keys.append(key)

        return keys

    def save(self) -> None:
        import tomli_w

        with open(self.config_path, "wb") as file:
            tomli_w.dump(self._config, file)

    def get(self, key: str) -> Optional[str]:
        if not self.profile:
            return None

        return self._config.get(self.profile, {}).get(key)

    def set(self, key: str, value: str) -> None:
        if not self.profile:
            raise NO_PROFILE_ERROR

        self._config[self.profile][key] = value

    def unset(self, key: str) -> None:
        if not self.profile:
            raise NO_PROFILE_ERROR

        self._config.get(self.profile, {}).pop(key, None)

    def get_internal(self, key: str) -> Optional[str]:
        if SETTINGS_SECTION not in self._config:
            self._config[SETTINGS_SECTION] = {}

        return self._config[SETTINGS_SECTION].get(key)

    def set_internal(self, key: str, value: Optional[str]) -> None:
        if SETTINGS_SECTION not in self._config:
            self._config[SETTINGS_SECTION] = {}

        if value is None:
            del self._config[SETTINGS_SECTION][key]
        else:
            self._config[SETTINGS_SECTION][key] = value

    def unset_internal(self, key: str) -> None:
        self._config.get(SETTINGS_SECTION, {}).pop(key, None)

    def delete_profile(self, profile: str) -> None:
        del self._config[profile]

    @contextmanager
    def edit(self) -> Iterator[Config]:
        if self._editing:
            # no-op
            yield self
        else:
            self._editing = True
            yield self
            self.save()
            self._editing = False



from pathlib import Path

CURR_DIR = Path(__file__).resolve().parent


def get_requirements():
    with open(CURR_DIR / "requirements.txt") as fp:
        requirements = fp.read().split()
    return requirements



from __future__ import annotations

from unittest import mock
from urllib.error import HTTPError
from urllib.request import Request

import pytest

from fal.toolkit.file.providers.fal import _maybe_retry_request

RETRY_CODES = [408, 409, 429, 500, 502, 503, 504]
NON_RETRY_CODES = [400, 401, 403, 404, 422]


class MockResponse:
    """Mock response object that mimics urllib.response.addinfourl"""

    def __init__(self, data: str = '{"result": "success"}', status: int = 200):
        self.data = data.encode()
        self.status = status
        self.headers = {"Content-Type": "application/json"}

    def read(self):
        return self.data

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def test_successful_request_no_retry():
    request = Request("https://example.com/test")

    with mock.patch("fal.toolkit.file.providers.fal._urlopen") as mock_urlopen:
        mock_response = MockResponse()
        mock_context = mock.MagicMock()
        mock_context.__enter__.return_value = mock_response
        mock_context.__exit__.return_value = None
        mock_urlopen.return_value = mock_context

        with _maybe_retry_request(request) as response:
            assert response == mock_response

        assert mock_urlopen.call_count == 1


@pytest.mark.parametrize("error_code", RETRY_CODES)
def test_retry_on_retryable_http_codes(error_code):
    request = Request("https://example.com/test")

    call_count = 0

    def mock_urlopen_side_effect(*args, **kwargs):
        nonlocal call_count
        call_count += 1

        if call_count == 1:
            raise HTTPError(
                url="https://example.com/test",
                code=error_code,
                msg=f"HTTP {error_code} Error",
                hdrs={},
                fp=None,
            )
        else:
            mock_response = MockResponse()
            mock_context = mock.MagicMock()
            mock_context.__enter__.return_value = mock_response
            mock_context.__exit__.return_value = None
            return mock_context

    with mock.patch(
        "fal.toolkit.file.providers.fal._urlopen", side_effect=mock_urlopen_side_effect
    ):
        with mock.patch(
            "fal.toolkit.utils.retry.time.sleep"
        ):  # Mock sleep to speed up tests
            with _maybe_retry_request(request) as response:
                assert response is not None

    assert call_count == 2


@pytest.mark.parametrize("error_code", NON_RETRY_CODES)
def test_no_retry_on_non_retryable_http_codes(error_code):
    request = Request("https://example.com/test")

    call_count = 0

    def mock_urlopen_side_effect(*args, **kwargs):
        nonlocal call_count
        call_count += 1
        raise HTTPError(
            url="https://example.com/test",
            code=error_code,
            msg=f"HTTP {error_code} Error",
            hdrs={},
            fp=None,
        )

    with mock.patch(
        "fal.toolkit.file.providers.fal._urlopen", side_effect=mock_urlopen_side_effect
    ):
        with pytest.raises(HTTPError) as exc_info:
            with _maybe_retry_request(request):
                pass

        assert exc_info.value.code == error_code
        assert call_count == 1


def test_non_http_exception_not_retried():
    """Test that non-HTTP exceptions are not retried"""
    request = Request("https://example.com/test")

    call_count = 0

    def mock_urlopen_side_effect(*args, **kwargs):
        nonlocal call_count
        call_count += 1
        raise ValueError("Network error")

    with mock.patch(
        "fal.toolkit.file.providers.fal._urlopen", side_effect=mock_urlopen_side_effect
    ):
        with pytest.raises(ValueError) as exc_info:
            with _maybe_retry_request(request):
                pass

        assert str(exc_info.value) == "Network error"
        assert call_count == 1


def test_max_retries_exhausted_for_retryable_errors():
    """Test that retries are exhausted after MAX_ATTEMPTS for retryable HTTP errors"""
    request = Request("https://example.com/test")

    call_count = 0

    def mock_urlopen_side_effect(*args, **kwargs):
        nonlocal call_count
        call_count += 1
        # Always fail with a retryable error
        raise HTTPError(
            url="https://example.com/test",
            code=500,
            msg="Internal Server Error",
            hdrs={},
            fp=None,
        )

    with mock.patch(
        "fal.toolkit.file.providers.fal._urlopen", side_effect=mock_urlopen_side_effect
    ):
        with mock.patch(
            "fal.toolkit.utils.retry.time.sleep"
        ):  # Mock sleep to speed up tests
            with pytest.raises(HTTPError) as exc_info:
                with _maybe_retry_request(request):
                    pass

        assert exc_info.value.code == 500
        assert call_count == 5
